#include "diskbrowse.h"
#include "diskbrowse.res"

DiskBrowser::DiskBrowser(QWidget *parent, 
			 const char *filter="*",
			 const char *name,
			 WFlags f) :
  QWidget(parent, name, f)
{
  QGridLayout *topLayout = new QGridLayout(this,10,4,5);

  // load the pixmaps we'll be using
  
  closedPixmap = new QPixmap((const char**)closed_xpm);
  openPixmap = new QPixmap((const char**)open_xpm);
  
  // set up child widgets
  
  dirTree = new QktsTreeList(this);
  dirTree->setExpandLevel(1);
  topLayout->addMultiCellWidget(dirTree,1,8,0,1);
  fileList = new QListBox(this);
  topLayout->addMultiCellWidget(fileList,1,8,2,3);
  connect(fileList,SIGNAL(selected(int)),this,SLOT(exportSelectedFile(int)));
  
  // when an item in the directory tree is highlighted, selected or 
  // expanded, notify us so we can do something
  
  connect(dirTree, SIGNAL(highlighted(int)), this, SLOT(dirHighlighted(int)));
  connect(dirTree, SIGNAL(expanded(int)), this, SLOT(dirExpanded(int)));
  connect(dirTree, SIGNAL(selected(int)), this, SLOT(dirSelected(int)));
  
  lblPath = new QLabel("/",this);
  topLayout->addMultiCellWidget(lblPath,0,0,0,3);
  lblFilt = new QLabel("Filter:",this);
  topLayout->addWidget(lblFilt,9,0);
  edtFilt = new QLineEdit(this);
  edtFilt->setText(filter);
  connect(edtFilt, SIGNAL(returnPressed()), this, SLOT(setFilter()));
  topLayout->addWidget(edtFilt,9,1);
  
  topLayout->activate();
  initializeTree();  
}

DiskBrowser::~DiskBrowser()
{
  // delete the pixmaps
  // child widgets will be deleted automatically
  
  delete closedPixmap;
  delete openPixmap;
}

// if we want it for some reason, i.e. status bar

uint DiskBrowser::fileCount()
{
  return fileList->count();
}
      
void DiskBrowser::dirExpanded(int index)
{
  advanceReadDirectories(index);
}

// here's where we change the folder pixmaps and list the
// files in the highlighted directory

void DiskBrowser::dirHighlighted(int index)
{
  static QDir dir;
  static QktsPath oldPath;
  QString dirPath;
  oldPath.setAutoDelete(TRUE);
  if(!oldPath.isEmpty()) {
    dirTree->changeItem(0, closedPixmap, &oldPath);
    oldPath.clear();
  }
  oldPath = *dirTree->itemPath(index);
  dirTree->changeItem(0, openPixmap, &oldPath);
  QktsPath pathCopy = oldPath;
  while(!pathCopy.isEmpty()) {
    dirPath.prepend(*pathCopy.pop());
    dirPath.prepend("/");
  }
  dir.setPath(dirPath);
  const QStrList *files = dir.entryList(edtFilt->text(),QDir::Files);
  if(files) {  
 		fileList->clear();
  	fileList->insertStrList(files);
  }
  lblPath->setText(dir.absPath());
}

void DiskBrowser::setFilter()
{
  dirHighlighted(dirTree->currentItem());
}

void DiskBrowser::dirSelected(int index)
{
 dirTree->expandOrCollapseItem(index);
}

// reads directories one level deeper in advance of tree expansion

void DiskBrowser::advanceReadDirectories(int index)
{
  QktsTreeListItem *parentItem = dirTree->itemAt(index);
  QktsPath *treePath = dirTree->itemPath(index);
  QktsTreeListItem *childItem = parentItem->getChild();
  while(childItem) {
    if(childItem->hasChild()) return; // already read!!!
    const char *childDir = childItem->getText();
    QktsPath treePathCopy = *treePath;
    QString filePath = childDir;
    while(!treePathCopy.isEmpty()) {
      filePath.prepend("/");
      filePath.prepend(*treePathCopy.pop());
    }
    treePath->push(new QString(childDir));
    QDir dir(filePath);
    const QStrList *dirs = dir.entryList(QDir::Dirs | QDir::Readable);
    if(dirs && dirs->count() != 2) { // prevent core dumps in /proc!
      QStrListIterator i(*dirs);
      for(; i.current(); ++i) {
	      QString currDir(i.current());
	      if(currDir == "." || currDir == "..")
	        continue;
	      dirTree->addChildItem(currDir, closedPixmap, treePath);
      }
    }
    childItem = childItem->getSibling();
    treePath->pop();
  }
  delete treePath;
}
      
// reads root directory and fills the tree list to two levels deep

void DiskBrowser::initializeTree()
{
  QDir rootDir = QDir::root();
  QString rootName = rootDir.absPath();
  dirTree->insertItem(rootName, closedPixmap, -1);
  QktsPath path;
  path.push(&rootName);
  const QStrList *dirs = rootDir.entryList(QDir::Dirs | QDir::Readable);
  if(dirs->count() == 2)
    return;
  QStrListIterator i(*dirs);
  for(; i.current(); ++i) {
    QString currDir(i.current());
    if(currDir == "." || currDir == "..")
      continue;
    dirTree->addChildItem(currDir, closedPixmap, &path);
    QDir subDir(rootDir.absFilePath(currDir));
    const QStrList *subDirs = subDir.entryList(QDir::Dirs | QDir::Readable);
    if(subDirs->count() == 2)
      continue;
    QStrListIterator j(*subDirs);
    path.push(new QString(currDir));
    for(; j.current(); ++j) {
      QString currSubDir(j.current());
      if(currSubDir != "." && currSubDir != "..")
        dirTree->addChildItem(currSubDir, closedPixmap, &path);
    }
    QString *junk = path.pop();
    delete junk;
  }
}

void DiskBrowser::exportSelectedFile(int i)
{
	emit selectedFile(i);
}

const char* DiskBrowser::getFileName(int i)
{
	static QString str;
	str = lblPath->text();
	if (i>fileList->count() || i<0)
		return NULL;
	else{
		str.append("/"); 
		str.append(fileList->text(i));
		return str.data();
	}
}

const char* DiskBrowser::getSelectedFileName()
{
	static QString str;
	str = lblPath->text();
	if (fileList->currentItem() < 0)
		return NULL;
	else{
		str.append("/");
		str.append(fileList->text(fileList->currentItem()));
		return str.data();
	}
}

const QStrList* DiskBrowser::getFileList()
{
	int i;
	static QStrList tmplist;
		
	tmplist.clear();
	for (i=0;i<fileList->count();i++){
		QString str = lblPath->text();
		str.append("/");
		str.append(fileList->text(i));
		tmplist.append(str);
	}
	return &tmplist;
}
