/* qktstreelist.h
-------------------------------------------------------------------
$Revision: 1.5 $
$Date: 1997/02/09 19:39:36 $
  
QktsTreeList class declaration

Copyright (C) 1996 Keith Brown and KtSoft

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABLILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details. You should have received a
copy of the GNU General Public License along with this program; if
not, write to the Free Software Foundation, Inc, 675 Mass Ave, 
Cambridge, MA 02139, USA.

-------------------------------------------------------------------
*/

#ifndef QktsTREE_LIST_H
#define QktsTREE_LIST_H

#include <qapp.h>       // used for QApplication::closingDown()
#include <qkeycode.h>           // used for keyboard interface
#include <qpainter.h>		// used to paint items
#include <qpixmap.h>		// used in items
#include <qstack.h>		// used to specify tree paths
#include <qstring.h>		// used in items
#include <qtablevw.h>		// base class for widget

// use stack of strings to represent path information
typedef QStack<QString> QktsPath;

class QktsTreeList;		// forward declaration

class QktsTreeListItem
{
  public:
    QktsTreeListItem(const char *theText = 0, 
		     const QPixmap *thePixmap = 0); // text can not be null when added to the list!
    virtual ~QktsTreeListItem();
    void appendChild(QktsTreeListItem *newChild);
    virtual QRect boundingRect(const QFontMetrics& fm) const;
    QktsTreeListItem *childAt(int index);
    uint childCount() const;
    int childIndex(QktsTreeListItem *child);
    bool drawExpandButton() const;
    bool drawText() const;
    bool drawTree() const;
    bool expandButtonClicked(const QPoint& coord) const;
    int getBranch() const;
    QktsTreeListItem *getChild();
    int getIndent() const;
    QktsTreeListItem *getParent();
    const QPixmap *getPixmap() const;
    QktsTreeListItem *getSibling();
    const char *getText() const;
    bool hasChild() const;
    bool hasParent() const;
    bool hasSibling() const;
    virtual int height(const QktsTreeList *theOwner) const;
    void insertChild(int index,
		     QktsTreeListItem *newItem);
    bool isExpanded() const;
    virtual void paint(QPainter *p, 
		       const QColorGroup& cg, 
		       bool highlighted);
    void removeChild(QktsTreeListItem *child);
    void setBranch(int level);
    void setChild(QktsTreeListItem *newChild);
    void setDrawExpandButton(bool doit);
    void setDrawText(bool doit);
    void setDrawTree(bool doit);
    void setExpanded(bool is);
    void setIndent(int value);
    void setParent(QktsTreeListItem *newParent);
    void setPixmap(const QPixmap *pm);
    void setSibling(QktsTreeListItem *newSibling);
    void setText(const char *t);
    virtual QRect textBoundingRect(const QFontMetrics& fm) const;
    virtual int width(const QktsTreeList *theOwner) const;
  protected:
    virtual int itemHeight(const QFontMetrics& fm) const;
    virtual int itemWidth(const QFontMetrics& fm) const;
  private:
    int branch;
    int indent;
    int numChildren;
    bool doExpandButton;
    bool doTree;
    bool doText;
    bool expanded;
    QRect expandButton;
    QktsTreeListItem *child;
    QktsTreeListItem *parent;
    QktsTreeListItem *sibling;
    QPixmap pixmap;
    QString text;
};

// easier declarations of function prototypes for forEvery type functions
typedef bool (QktsTreeList::*QktsForEveryM)
    (QktsTreeListItem *, void *);
typedef bool (*QktsForEvery)
    (QktsTreeListItem *, void *);

struct QktsItemSearchInfo {
  int count;
  int index;
  QktsTreeListItem *foundItem;
  QktsItemSearchInfo() : count(0), index(0), foundItem(0) {}
};

class QktsTreeList : public QTableView
{
  Q_OBJECT
  public:
    QktsTreeList(QWidget *parent = 0, 
		             const char *name = 0,
		             WFlags f = 0);
    virtual ~QktsTreeList();
    void addChildItem(const char *theText,
                      const QPixmap *thePixmap,
                      int index);
    void addChildItem(const char *theText,
                      const QPixmap *thePixmap,
                      const QktsPath *thePath); 
    void addChildItem(QktsTreeListItem *newItem,
                      int index);
    void addChildItem(QktsTreeListItem *newItem,
                      const QktsPath *thePath);                                                             
    bool autoBottomScrollBar() const;
    bool autoScrollBar() const;
    bool autoUpdate() const;
    bool bottomScrollBar() const;
    void changeItem(const char *newText, 
		    const QPixmap *newPixmap, 
		    int index);
    void changeItem(const char *newText,
		    const QPixmap *newPixmap,
		    const QktsPath *thePath);
    void clear();
    uint count();
    int currentItem() const;
    void collapseItem(int index);
    void expandItem(int index);
		int expandLevel() const;
    void expandOrCollapseItem(int index);
    void forEveryItem(QktsForEvery func, 
		      void *user);
    void forEveryVisibleItem(QktsForEvery func, 
			     void *user);
    QktsTreeListItem *getCurrentItem();
    int indentSpacing();
    void insertItem(const char *theText,
		    const QPixmap *thePixmap,
		    int index = -1,
		    bool prefix = TRUE);
    void insertItem(const char *theText,
		    const QPixmap *thePixmap,
		    const QktsPath *thePath,
		    bool prefix = TRUE);
    void insertItem(QktsTreeListItem *newItem, 
		    int index = -1,
		    bool prefix = TRUE); 
    void insertItem(QktsTreeListItem *newItem,
		    const QktsPath *thePath,
		    bool prefix = TRUE);
    QktsTreeListItem *itemAt(int index);
    QktsTreeListItem *itemAt(const QktsPath *path);
    int itemIndex(QktsTreeListItem *item);
    QktsPath *itemPath(int index);
    void join(int index);
    void join(const QktsPath *path);
    void lowerItem(int index);
    void lowerItem(const QktsPath *path);
    void raiseItem(int index);
    void raiseItem(const QktsPath *path);
    void removeItem(int index);
    void removeItem(const QktsPath *thePath);
    bool scrollBar() const;
    void setAutoUpdate(bool enable);
    void setBottomScrollBar(bool enable);
    void setCurrentItem(int index);
    void setExpandButtonDrawing(bool enable);
    void setExpandLevel(int level);
    void setIndentSpacing(int spacing);
    void setScrollBar(bool enable);
    void setShowItemText(bool enable);
    void setSmoothScrolling(bool enable);
    void setTreeDrawing(bool enable);
    bool showItemText() const;
    bool smoothScrolling() const;
    void split(int index);
    void split(const QktsPath *path);
    QktsTreeListItem *takeItem(int index);
    QktsTreeListItem *takeItem(const QktsPath *path);
    bool treeDrawing() const;
    int visibleCount();
  signals:
    void collapsed(int index);
    void expanded(int index);
    void highlighted(int index);
    void selected(int index);
  protected:
    void addChildItem(QktsTreeListItem *theParent,
		      QktsTreeListItem *theChild);
    virtual int cellHeight(int row);
    void changeItem(QktsTreeListItem *toChange,
		    int itemIndex,
		    const char *newText,
		    const QPixmap *newPixmap);
    bool checkItemPath(const QktsPath *path) const;
    bool checkItemText(const char *text) const;
    void collapseSubTree(QktsTreeListItem *subRoot);
    bool countItem(QktsTreeListItem *item,
		   void *total);
    void expandOrCollapse(QktsTreeListItem *parentItem);
    void expandSubTree(QktsTreeListItem *subRoot);
    bool findItemAt(QktsTreeListItem *item,
		    void *user);
		void fixChildBranches(QktsTreeListItem *parentItem);
    virtual void focusInEvent(QFocusEvent *e);
    void forEveryItem(QktsForEveryM func, 
		      void *user);
    void forEveryVisibleItem(QktsForEveryM func,
			     void *user);
    bool getItemIndex(QktsTreeListItem *item,
		      void *user);
    bool getMaxItemWidth(QktsTreeListItem *item,
			 void *user);
    void insertItem(QktsTreeListItem *referenceItem,
		    QktsTreeListItem *newItem,
		    bool prefix);
		void join(QktsTreeListItem *item);
    virtual void keyPressEvent(QKeyEvent *e);
    void lowerItem(QktsTreeListItem *item);
    virtual void mouseDoubleClickEvent(QMouseEvent *e);
    virtual void mouseMoveEvent(QMouseEvent *e);
    virtual void mousePressEvent(QMouseEvent *e);
    virtual void mouseReleaseEvent(QMouseEvent *e);
    virtual void paintCell(QPainter *p, int row, int col);
    virtual void paintHighlight(QPainter *p, 
      QktsTreeListItem *item);
    virtual void paintItem(QPainter *p, QktsTreeListItem *item, 
      bool highlighted);
    void raiseItem(QktsTreeListItem *item);
    QktsTreeListItem *recursiveFind(QktsTreeListItem *subRoot,
				    QktsPath *path);
    bool setItemExpanded(QktsTreeListItem *item, void *);
    bool setItemExpandButtonDrawing(QktsTreeListItem *item, void *);
    bool setItemIndent(QktsTreeListItem *item, void *);
    bool setItemShowText(QktsTreeListItem *item, void *);
    bool setItemTreeDrawing(QktsTreeListItem *item, void *);
    void split(QktsTreeListItem *item);
    void takeItem(QktsTreeListItem *item);
    virtual void updateCellWidth();
    QktsTreeListItem *treeRoot;
    bool clearing;
    int current;
    bool drawExpandButton;
    bool drawTree;
    int expansion;
    bool goingDown;
    int indent;
    int maxItemWidth;
    bool showText;
};

#endif // QktsTREE_LIST_H
